import socket
import datetime
from future.moves.urllib.request import urlopen
from future.moves.urllib.error import URLError
import os
from cestel_logging import log, log_exception
import logging

from lxml import etree
from generic_module import Module

class PapagoModule(Module):
  def __init__(self, name, udp_ip, udp_port, device_ip, interval, indexing_start, root_path, relevant_for_listeners):
    Module.__init__(self, name, root_path)
    self.udp_ip = udp_ip
    self.udp_port = udp_port
    
    self.interval = interval
    
    self.device_ip = device_ip
    
    self.protocol = "http://"
    self.loc = "/fresh.xml"
    
    self.indexing_start = indexing_start
    
    #don't support (for now)
    self.relevant_for_listeners = relevant_for_listeners #t, 0
    
  def generate_path(self):
    pth = self.sites_path
    pth = pth + "/" + self.sname
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/ext"
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/temperature"
    if not os.path.exists(pth):
      os.mkdir(pth)
    #pth = pth + "/papago"
    #if not os.path.exists(pth):
      #os.mkdir(pth)
    """
    pth = pth + "/" + self.typ
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + self.name 
    if not os.path.exists(pth):
      os.mkdir(pth)
    """
    #print pth
    return pth
  
  def get_data(self):
    data_url = self.protocol + self.device_ip + self.loc
    log(self.logger, 'Data url {0}'.format(data_url), sev=logging.DEBUG)
    try:
      request = urlopen(data_url)
    except URLError as e:
      log(self.logger, 'get_data() failed: {0}.'.format(e), sev=logging.WARNING)
      return 'Err'
    except:
      log_exception(self.logger, 'get_data error:')
      return "Err"
    xml = request.read()   
    xmlns = xml[xml.find("xmlns=\""):xml.find("\">") + 1]
    xml = xml.replace(xmlns, "")
    
    temp_data = etree.fromstring(xml)
    
    temperature = etree.Element("temperature_reading")
    mname = etree.Element("module_name")
    mname.text = self.name
    
    #print etree.tostring(temp_data)
    #ts = temp_data.find("status").attrib["time"]
    our_ts = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
    ts_el = etree.Element("ts")
    ts_el.text = our_ts
    temperature.append(ts_el)
    
    values = etree.Element("values")
    temperature.append(values)
    
    sns = temp_data.findall("sns")
    
    curr_idx = self.indexing_start
    
    for sn in sns:
      value = float(sn.attrib["val"])
      status = sn.attrib["status"]
      ut = int(sn.attrib["unit"])
      if status != "0":
        log(self.logger, 'Status: {0}. Consult documentation.'.format(status))
      #sensor = etree.Element("sensor")
      
      if ut == 1:
        value = (value - 32.0) * (5 / 9) 
      if ut == 2:
        value = value - 273.15
      
      val = etree.Element("value")
      val.text = str(value)
      
      val.attrib["idx"] = str(curr_idx)
      #maybe usable at a later time
      val.attrib["name"] = "some_name"
      val.attrib["unit"] = "C"
      values.append(val)
      curr_idx+=1
    return temperature
  
  def save_to_file(self, data):
    if self.sname == "":
      return
    first_line = "<td><site><name>" + self.sname + "</name><temperature_readings>\n"
    last_line = "</temperature_readings></site></td>"
    try:
      date = datetime.datetime.now().strftime('%Y-%m-%d')
      path = self.generate_path()
      fd = open(path + "/" + date + ".xml", "a+")
      #remove last line          
      fd.seek(0, os.SEEK_END)
      pos = fd.tell() - 1
      while pos > 0 and fd.read(1) != "\n":
        pos -= 1
        fd.seek(pos, os.SEEK_SET)
      if pos > 0:
        fd.seek(pos, os.SEEK_SET)
        fd.truncate()
      #if file is empty, add first line
      if pos == fd.tell() - 1:
        fd.write(first_line + etree.tostring(data).decode() + "\n" + last_line)
      else:
        fd.write(etree.tostring(data).decode() + "\n" + last_line)
      fd.close()
    except:
      try:
        fd.close()
      except:
        log(self.logger, 'Failed to close:', sev=logging.DEBUG, exc_info=True)
      log_exception(self.logger, 'Failed to close:')
  
  def send_data(self, data):
    conf_str_data = self.xml_to_conf_str("mydevice", data)
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
      sock.sendto(conf_str_data, (self.udp_ip, self.udp_port))
    finally:
      sock.close()
  
  def xml_to_conf_str(self, section, data):
    values = data.find("values").findall("value")
    conf_str = "[" + section + "]\n"
    for value in values:
      conf_str+="T" + value.attrib["idx"] + "__" + value.attrib["unit"] + "=" + value.text + "\n"                  
    return conf_str
    
  def run(self):
    self.alive = True
    self.end = False
    
    while True:
      if self.end:
        self.alive = False
        log(self.logger, 'Thread closed correctly.', sev=logging.DEBUG)
        return
      data = self.get_data()
      if data != "Err":
        self.send_data(data)
        self.save_to_file(data)
      self.zzzzz(self.interval)
