import datetime
import sys
from sys import platform
import socket
import re
import configparser
import os
import subprocess
import pefile
from cestel_logging import log, log_exception
import logging
import requests

if platform == "win32":
  import winreg

from datetime import timedelta
from lxml import etree
from generic_module import Module
 
class StatusModule(Module):
  def __init__(self, name, host, time_interval, version, url, protocol, root_path, ntp_conf_path, ovpn_path, save_curr_to_disk):
    Module.__init__(self, name, root_path)
    self.time_interval = time_interval
    self.host = host
    self.version = version
    self.url = url
    self.protocol = protocol 
    self.ntp_conf_path = ntp_conf_path
    self.ovpn_path = ovpn_path
    self.save_curr_to_disk = save_curr_to_disk
    
    self.ctu_fw_version = ""
    self.hist_fw_version = ""
    self.cat_fw_version = ""
    self.caz_fw_version = ""
    self.voltage = ""
    
    self.mcp_mod_time = 0
    self.mcp_version = ""
    
  def set_voltage(self, voltage):
    #self.ctu_fw_version = fw_version
    self.voltage = voltage
    
  def set_ctu_fw(self, fw_version):
    self.ctu_fw_version = fw_version
    #self.voltage = voltage
    
  def set_hist_fw(self, fw_version):
    self.hist_fw_version = fw_version
    
  def set_caz_fw(self, fw_version):
    self.caz_fw_version = fw_version
    
  def set_cat_fw(self, fw_version):
    self.cat_fw_version = fw_version
    #self.voltage = voltage

  def swmCurSite(self, loc):
    try:
      cfg = configparser.RawConfigParser()
      cfg.read(loc)
      #cfg.read('d:/siwim_mkiii/siwim_e/conf/siwim.conf')
      return cfg.get('global', 'site')
    except:
      return 'unknown'
    
  def w32ovpn_version(self):  
    sp = subprocess.Popen([self.ovpn_path, "--version"], stdout=subprocess.PIPE)
    output = sp.communicate()[0]
    return output.split(b" ")[1].decode()  # Output of the application is binary and thus has to be split by "binary spaces". We could also decode before splitting, but that's probably redundant.
  """
  def uniX32ovpn_version(self):  
    sp = subprocess.Popen(["dpkg", "-l", "openvpn"], stdout=subprocess.PIPE)
    output = sp.communicate()[0]
    lines = output.split("\n")
    #log(str(lines))
    ver = lines[len(lines) - 2].split()[2]
    return ver
  """  
  #def uniXswmCurSite(self):
    #try:
      #cfg = C  onfigParser.RawConfigParser()
      #cfg.read('/siwim/siwim_e/conf/siwim.conf')
      #return cfg.get('global', 'site')
    #except:
      #return 'unknown'
  
  def getUniXhostname(self):
    sp = subprocess.Popen(["hostname"], stdout=subprocess.PIPE)
    output = sp.communicate()[0]
    return output.strip()

  def get_file_info(self, filename):
    # Version is packed into 2 integers, with each containing 2 values. Thus these convenience functions return least/most significant words.
    def lsw(dword):
      return dword & 0x0000ffff
    def msw(dword):
      return dword >> 16

    pe = pefile.PE(filename)
    ms = pe.VS_FIXEDFILEINFO[0].ProductVersionMS
    ls = pe.VS_FIXEDFILEINFO[0].ProductVersionLS
    return '{0}.{1}.{2}.'.format(msw(ms), lsw(ms), msw(ls))

  def get_registry_value(self, key, subkey, value):
    key = getattr(_winreg, key)
    handle = winreg.OpenKey(key, subkey)
    (value, type) = winreg.QueryValueEx(handle, value)
    return value

  def _cpu(self):
    try:
      return self.get_registry_value("HKEY_LOCAL_MACHINE", "HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0", "ProcessorNameString")
    except:
      return self.get_registry_value("HKEY_LOCAL_MACHINE", "HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0", "Identifier")

  def _mhz(self):
    return self.get_registry_value("HKEY_LOCAL_MACHINE", "HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0", "~MHz")

  def _sysver(self):
    return 'MK3'

  #class MEMORYSTATUS(Structure):
    #_fields_ = [('dwLength', DWORD), ('dwMemoryLoad', DWORD), ('dwTotalPhys', DWORD), ('dwAvailPhys', DWORD), ('dwTotalPageFile', DWORD), ('dwAvailPageFile', DWORD), ('dwTotalVirtual', DWORD), ('dwAvailVirtual', DWORD),]

  #def winmem(self):
    #ms = self.MEMORYSTATUS()
    #windll.kernel32.GlobalMemoryStatus(byref(ms))
    #return ms

  def gpsLoc(self):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(5.0)
    data = 'No GPS data'
    try:
      s.connect(('192.168.4.100', 9494))
      data = s.recv(1024).decode()
      s.close()
    except:
      log(self.logger, 'Obtaining legacy SierraGpsLoc failed', sev=logging.DEBUG, exc_info=True)

    p = re.compile('GPGGA,([.\d]*),([.\d]*),([NS]),([.\d]*),([EW])')
    a = p.search(data)

    if a == None:
      return (0, 0)
    else:
      latd = float(a.group(2)[:2])
      lond = float(a.group(4)[:3])
      latm = float(a.group(2)[2:])
      lonm = float(a.group(4)[3:])
      
      lat = latd + (latm / 60)
      lon = lond + (lonm / 60)

      if a.group(3) != 'N':
        lat = -lat
      if a.group(5) != 'E':
        lon = -lon
      return (lat, lon)

  def _os_version(self):
    os = self.get_registry_value("HKEY_LOCAL_MACHINE","SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion","ProductName")
    try:
      sp = self.get_registry_value("HKEY_LOCAL_MACHINE","SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion","CSDVersion")
    except:
      sp = 'Vanilla'
    build = self.get_registry_value("HKEY_LOCAL_MACHINE","SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion","CurrentBuildNumber")
    return "%s %s (build %s)" % (os, sp, build)

  def _recvfind(self, sock, n, to_sec=5):
    d = ''
    absto = datetime.datetime.now() + timedelta(seconds=to_sec)
    while (-1 == d.find(n)) and (datetime.datetime.now() < absto):
      d += sock.recv(256).decode(errors='replace')
    return re.sub(r'[^a-zA-Z0-9 \n]', '', d)

  def SierraImsi(self, pwd):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(5.0)
    imsi = 'No IMSI'
    try:
      s.connect(('192.168.4.100', 2332))
      data = self._recvfind(s, ':')
      if data.find('login') >= 0:
        #print 'GX'
        s.sendall('user\r\n'.encode())
        data = self._recvfind(s, ':')
        #print 'MP'
      s.sendall((pwd + '\r\n').encode())
      data = self._recvfind(s, 'OK')
      s.sendall('at+cimi?\r\n'.encode())
      data = self._recvfind(s, 'OK')
      data = data.split()
      if len(data) > 1: imsi = data[-2]
      s.close()
    except:
      log(self.logger, 'Obtaining legacy SierraImsi failed', sev=logging.DEBUG, exc_info=True)
    return imsi

  def swmiver(self):
    return ".".join(str(x) for x in self.version)

  def ntpsvr(self, loc):
    fn = loc
    if not os.path.exists(fn):
      return 'missing'
    f = open(fn)
    try:
      lines = f.readlines()
    finally:
      f.close()
    for l in lines:
      if l.strip().startswith('server'):
        return l.strip().split()[1] 
    for l in lines:
      if l.strip().startswith('pool'):
        return l.strip().split()[1] 
    return 'unknown'

  def swmstat(self):
    kvp = {}
    try:
      kvp['python_version']      = str(sys.version_info[0])+'.'+str(sys.version_info[1])+'.'+str(sys.version_info[2])
      kvp['siwim_version'] = 'MKIII' if sys.version_info[0] == 2 else 'MKIV'
    except:
      log(self.logger, 'Python version not saved.', sev=logging.WARNING)
    try:
      kvp['timestamp']  = datetime.datetime.now().strftime('%Y-%m-%d %H:%M')  
    except:
      log(self.logger, 'Timestamp not saved.', sev=logging.WARNING)
    try:
      if platform == "win32":
        kvp['computer_name']   = os.getenv('COMPUTERNAME')
      else:
        kvp['computer_name']   = self.getUniXhostname()
    except:
      log(self.logger, 'Computer name not saved.', sev=logging.WARNING)
    # To reduce repeated code, map values of partitions depending on platform and save them to partitions dict.
    if platform == 'win32':
      partitions = {'disk_c': 'c:', 'disk_d': 'd:', 'disk_e': 'e:'}
    else:
      partitions = {'disk_c': '/', 'disk_d': self.root_path, 'disk_e': '/'}
    for key, p in partitions.items():
      try:
        kvp['{0}_free'.format(key)] = self.get_free_mb(p)
      except:
        log(self.logger, 'Device {0} not found.'.format(p), sev=logging.WARNING)
    try:
      if platform == "win32":
        kvp['siwim_site_name']       = self.swmCurSite(self.root_path + "/siwim_e/conf/siwim.conf")
      else:
        kvp['siwim_site_name']       = self.swmCurSite(self.root_path + "/siwim_e/conf/siwim.conf")
    except:
      log(self.logger, 'Site name not saved.', sev=logging.WARNING)
    try:
      kvp['siwim_system_voltage']       = self.swmVoltageCTU()
    except:
      log(self.logger, 'Voltage not saved.', sev=logging.WARNING)
    try:
      kvp['siwim_e_version'] = self.get_file_info(self.root_path + '/siwim_e/bin/siwim_mcp.exe')
    except:
      log(self.logger, 'E version not saved.', sev=logging.WARNING)
    try:
      kvp['siwim_i_version']    = self.swmiver()
    except:
      log(self.logger, 'I version not saved.', sev=logging.WARNING)
    try:
      (lat,lon) = self.gpsLoc()
      kvp['gps_lat']        = lat
      kvp['gps_lon']        = lon
    except:
      log(self.logger, 'GPS not saved.', sev=logging.WARNING)
    for key, p in partitions.items():
      try:
        kvp['{0}_total'.format(key)] = self.get_total_mb(p)
      except:
        log(self.logger, 'Total MB not saved for {0}.'.format(p), sev=logging.WARNING)
    try:
      kvp['gsm_imsi']       = self.SierraImsi("12345")
      if kvp['gsm_imsi'] == 'No IMSI':
        kvp['gsm_imsi']       = self.SierraImsi("siwim323")
    except:
      log(self.logger, 'GSM imsi not saved.', sev=logging.WARNING)
    try:
      if self.ntp_conf_path == "" or self.ntp_conf_path is None:  # FIXME I got an error where path was none rather than empty string, so probably worth checking
        if platform == "win32":
          kvp['ntp_server']        = self.ntpsvr("c:/windows/system32/drivers/etc/ntp.conf") 
        else:
          kvp['ntp_server']        = self.ntpsvr("/etc/ntp.conf") 
      else:
        kvp['ntp_server']        = self.ntpsvr(self.ntp_conf_path) 
    except:
      log(self.logger, 'NTP server not saved', sev=logging.WARNING)
    try:
      if self.ctu_fw_version != "":
        kvp['ctu_firmware_version']        = self.ctu_fw_version
    except:
      log(self.logger, 'CTU not saved.', sev=logging.WARNING)
    try:
      if self.hist_fw_version != "":
        kvp['hist_firmware_version']        = self.hist_fw_version
    except:
      log(self.logger, 'Hysteresis not saved.', sev=logging.WARNING)
    try:
      if self.cat_fw_version != "":
        kvp['cat_firmware_version']        = self.cat_fw_version
    except:
      log(self.logger, 'Cat not saved.', sev=logging.WARNING)
    try:
      if self.caz_fw_version != "":
        kvp['caz_firmware_version']        = self.caz_fw_version
    except:
      log(self.logger, 'Caz not saved.', logging.WARNING)
    try:
      #if platform == "win32":
      kvp['openvpn_version']     = self.w32ovpn_version()
      #else:
        #kvp['openvpn_version']     = self.uniX32ovpn_version()
    except:
      log(self.logger, 'OpenVPN not saved.', logging.WARNING)
    
    for key, val in self.upstream_info.items():
      try:
        kvp[key[1]]     = val
      except:
        log(self.logger, key[1] + ' not saved.', logging.WARNING)
    return kvp
    
  def swmVoltageCTU(self):
    if self.voltage != "":
      return str(self.voltage)
    else:
      fn = self.root_path + '/siwim_e/log/ctu_' + datetime.datetime.now().strftime('%Y-%m-%d') + '.txt'
      if not os.path.exists(fn):
        return '?'
      try:
        f = open(fn,'r')
        return f.readlines()[-1].split('\t')[1][:4]
      finally:
        f.close()
    
  def run(self):
    self.alive = True
    self.end = False
    try:
      while True:
        if self.end:
          self.alive = False
          log(self.logger, 'Thread closed correctly.', sev=logging.DEBUG)
          return
        try:
          kvp = self.swmstat()
          root = etree.Element("swmstat") 
          for key, val in kvp.items():
            node = etree.Element(key) 
            node.text = str(val)
            root.append(node)
          if self.save_curr_to_disk == 1:
            log(self.logger, 'Writing status to disk ...')
            fd = open("log/swmstat.xml", "w")
            fd.write(etree.tostring(root).decode())
            fd.close()
          resp = requests.post('{0}://{1}{2}'.format(self.protocol, self.host, self.url), {'sys': kvp["computer_name"].lower(), 'msg': etree.tostring(root).decode()}, timeout=10, verify=False)
          if resp.status_code != 204:  # Log, if response isn't empty.
            log(self.logger, 'Response: {0}'.format(resp.text), sev=logging.WARNING)
        except socket.error as e:
          log(self.logger, 'Connection attempt failed ({0}).'.format(e), sev=logging.WARNING)
        except:
          log_exception(self.logger, 'General error:')
        #time.sleep(self.time_interval)
        self.zzzzz(self.time_interval)
    except:
      log_exception(self.logger, 'Fatal error:')
      self.alive = False
      